/* SPDX-License-Identifier: Zlib */

#ifndef GIRARA_H
#define GIRARA_H

#include "macros.h"

#include "datastructures.h"
#include "input-history.h"
#include "log.h"
#include "template.h"
#include "utils.h"

#endif
