/* SPDX-License-Identifier: Zlib */

#ifndef GIRARA_INTERNAL_H
#define GIRARA_INTERNAL_H

#include <glib.h>

#include "types.h"
#include "macros.h"

int list_strcmp(const void* data1, const void* data2);

#endif
